#include "matrix.h"

namespace std {

// ******************* CONSTRUCTORS *******************

// constructor
matrix::matrix(unsigned int rowNo, unsigned int colNo, T v) {
    if (!rowNo || !colNo)
        error("ERROR in matrix::matrix() : Zero matrix dimension!");
        
    this->rowNo = rowNo;
    this->colNo = colNo;
    permutationParity = 1;

    vals = new vector<vector<T> >;
    vals->resize(rowNo);
	for(unsigned int r=0; r<=rowNo-1; ++r)
		(*vals)[r].resize(colNo,v);

	luMatrix = NULL;
    prmVector = NULL;
    sclVector = NULL;
}

// copy constructor
matrix::matrix(const matrix& m) {
    rowNo = m.rowNo;
    colNo = m.colNo;

    vals = new vector<vector<T> >;
    *vals = *m.vals;    // here the operator= of the STL vector class is used
    
    copyLUMatrices(m);
}


// ******************* DESTRUCTOR *******************

// destructor
matrix::~matrix() {
	delete vals;
    deleteLUMatrices();
}


// ******************* FEW USEFUL FUNCTIONS *******************

// transpose function (modify & return *this object)
matrix & matrix::transpose() {
    matrix tmp(colNo,rowNo);

    for(unsigned int r=0; r<=rowNo-1; ++r)
        for(unsigned int c=0; c<=colNo-1; ++c)
	        tmp(c,r) = (*vals)[r][c];
    (*this) = tmp;

    return *this;
}

// randomise the matrix (modify & return *this object)
matrix & matrix::randomize(unsigned int limit) {
    srand(time(NULL));
    for(unsigned int r=0; r<=rowNo-1; ++r)
        for(unsigned int c=0; c<=colNo-1; ++c)
	        (*vals)[r][c] = rand() % (limit+1);
    deleteLUMatrices();

    return *this;
}


// ******************* SELF-MODIFYING OPERATORS *******************

// assignment operator
matrix& matrix::operator= (const matrix& m) {
    rowNo = m.rowNo;
    colNo = m.colNo;
    *vals = *m.vals;    // here the operator= of the STL vector class is used

    deleteLUMatrices();
    copyLUMatrices(m);

    return *this;
}

// addition assignment operator
matrix& matrix::operator+= (const matrix& m) {
    if (rowNo!=m.rowNo || colNo!=m.colNo)
        error("ERROR in matrix::operator+= : Inconsistent matrix sizes!");

    for(unsigned int r=0; r<=rowNo-1; ++r)
        for(unsigned int c=0; c<=colNo-1; ++c)
	       (*vals)[r][c] += m(r,c);

    deleteLUMatrices();

    return *this;
}

// matrix multiplication assignment operator
matrix& matrix::operator*= (const matrix& m) {
    if (colNo != m.rowNo)
        error("ERROR in matrix::operator*= : Inconsistent matrix sizes!");

    matrix tmp(rowNo,m.colNo);

    for (unsigned int r=0; r<=rowNo-1; ++r)
        for (unsigned int c=0; c<=m.colNo-1; ++c) {
            tmp(r,c) = 0;
            for(unsigned int i=0; i<=colNo-1; ++i)
                tmp(r,c) += (*vals)[r][i]*m(i,c);
        }
    *this = tmp;

    return *this;
}

// scalar multiplication assignment operator
matrix& matrix::operator*= (const T& v) {
    for(unsigned int r=0; r<=rowNo-1; ++r)
	    for(unsigned int c=0; c<=colNo-1; ++c)
	       (*vals)[r][c] *= v;
    if (luMatrix) {
        if (v != 0) {
            for(unsigned int r=0; r<=rowNo-1; ++r)
                (*sclVector)[r] /= v;
        } else deleteLUMatrices();
    }

    return *this;
}


// ******************* SUBSCRIPT OPERATORS *******************

// subscript operator to make individual elements available
T& matrix::operator() (unsigned int r, unsigned int c) {
    if (r>=rowNo || c>=colNo)
        error("ERROR in matrix::operator(): Index out of range!");
    deleteLUMatrices();
    
    return (*vals)[r][c];
}

// subscript operator to read individual elements (access only)
T matrix::operator()(unsigned int r, unsigned int c) const {
    if (r>=rowNo || c>=colNo)
        error("ERROR in matrix::operator[]: Index out of range!");

    return (*vals)[r][c];
}


// ******************* ACCESS to LU-RELATED DATA *******************

// return the r-th element of the permutation vector
unsigned int matrix::getLUPermutationVector(unsigned int r) {
    if (r>=rowNo)
        error("ERROR in matrix::getLUPermutationVector: Index out of range!");

    if (luMatrix==NULL)
        luDecompose();

    return (*prmVector)[r];
}

// return the r-th element of the scaling vector
T matrix::getLUScalingVector(unsigned int r) {
    if (r>=rowNo)
        error("ERROR in matrix::getLUScalingVector: Index out of range!");

    if (luMatrix==NULL)
        luDecompose();

    return (*sclVector)[r];
}

// return the permutation vector 
matrix matrix::getLUPermutationVector(void) {
    if (luMatrix==NULL)
        luDecompose();

    matrix tmp(rowNo,1);

    for(unsigned int r=0; r<=rowNo-1; ++r)
        tmp(r,0) = (*prmVector)[r];

    return tmp;
}

// return the scaling matrix
matrix matrix::getLUScalingVector(void) {
    if (luMatrix==NULL)
        luDecompose();

    matrix tmp(rowNo,1);

    for(unsigned int r=0; r<=rowNo-1; ++r)
        tmp(r,0) = (*sclVector)[r];

    return tmp;
}

// return the L matrix
matrix matrix::getLMatrix(void) {
    if (luMatrix==NULL)
        luDecompose();

    matrix tmp(rowNo,colNo);    // by the default argument filled with zeros

    for(unsigned int r=0; r<=rowNo-1; ++r)
	    for(unsigned int c=0; c<=r; ++c)   // only the lower part
            tmp(r,c) = (r==c) ? 1 : (*luMatrix)[r][c];

    return tmp;
}

// return the U matrix
matrix matrix::getUMatrix(void) {
    if (luMatrix==NULL)
        luDecompose();

    matrix tmp(rowNo,colNo);    // by the default argument filled with zeros

    for(unsigned int r=0; r<=rowNo-1; ++r)
	    for(unsigned int c=r; c<=colNo-1; ++c) // only the upper part
            tmp(r,c) = (*luMatrix)[r][c];

    return tmp;
}


// ******************* LS SOLVING *******************
matrix matrix::solveLS(const matrix& b) {
// delete the line below, when ready
    return matrix(b.getRowNo(),b.getColNo());

// Check if the dimensions (row no) of the vector b and the *this matrix comply
// and if b is really a vector (b.colNo == 1).
// If not, call the error() function, i.e. quit the program

// Check if the *this matrix is already decomposed. If not, call luDecompose().

// Define matrices b1 (to store rescaled and permutated b),
// y (to store the solution of Ly=b), x (to store the solution of Ux=y).

// Define & initialise a dummy variable to store intermediate sums 
// during forward/back substitutions.

// rescale and permutate the vector b, store the result in the vector b1.
// Notice that the rescaling is done BEFORE the permutation (see luDecompose()),
// not the opposite.

// Do the forward substitution to solve Ly=b1, do not forget to initialise

// Do the back substitution to solve Ux=y
    
// return x;
}


// ******************* THE DETERMINANT *******************
T matrix::det(void) {
// delete the line below, when ready
    return 0;

// Check if the *this matrix is square.
// If not, call the error() function, i.e. quit the program

// Check if the *this matrix is already decomposed. If not, call luDecompose().

// Define & initialise a dummy variable to store intermediate products.

// Compute the determinant. Do not forget about the permutation parity 
// (permutationParity) and the scaling information (sclVector)
        
// Return the dummy variable containing the determinant
}


// ******************* INTERNAL (PRIVATE) FUNCTIONS *******************

void matrix::error(const char* pMsg) const {
    cerr <<pMsg <<endl <<endl;
    system("pause");    // for Dev C++ users only
    exit(1);
}


// ******************* LU DECOMPOSITION *******************

// do LU decomposition
void matrix::luDecompose(void) {
    if (rowNo != colNo)
        error("ERROR in matrix::luDecompose: Matrix not square!");

    unsigned int k,i,j;
    T tmp1,tmp2;

    initialiseLUMatrices();
    luRescale();

    for(k=0; k<=rowNo-2; ++k) {            // for each row ...
        tmp1 = fabs((*luMatrix)[k][k]);        
	    for(j=k, i=k+1; i<=rowNo-1; ++i) { // go through all the rows below the k-th row
	        tmp2=fabs((*luMatrix)[i][k]);     
            if (tmp2 > tmp1) {             // and check if pivoting is necessary
                j = i;                     // if yes, store the row number
                tmp1 = tmp2;
            }
        }
        if (tmp1==0)
            error("ERROR in matrix::luDecompose: Matrix is singular!");
        if (j != k)                        // swap the rows, if needed
            luSwapRows(j,k);
	    for(i=k+1; i<=rowNo-1; ++i) {      // do the elimination
            tmp1 = (*luMatrix)[i][k]/(*luMatrix)[k][k];
	        for(j=k+1; j<=colNo-1; ++j)
                (*luMatrix)[i][j] -= tmp1*(*luMatrix)[k][j];
            (*luMatrix)[i][k] = tmp1;
        }
    }
}

// initialise prmVector, 
// fill sclVector and rescale the matrix copying it to luMatrix
// rescale to keep the sum of the absolute values in each row constant (1)
void matrix::luRescale(void) {
    unsigned int i;
    T sum;

    for(unsigned int k=0; k<=rowNo-1; ++k) {
        (*prmVector)[k] = k;
	    for(sum=0, i=0; i<=colNo-1; ++i)
            sum += fabs((*vals)[k][i]);
        if (sum==0)
            error("ERROR in matrix::luRescale: Matrix is singular!");
        else (*sclVector)[k] = 1/sum;
	    for(i=0; i<=colNo-1; ++i)
            (*luMatrix)[k][i] = (*vals)[k][i]/sum;
    }
}

void matrix::luSwapRows(unsigned int r1, unsigned int r2) {
    vector<T> * tmpM = new vector<T>;   
    unsigned int tmp;

    *tmpM = (*luMatrix)[r1];                // row swapping
    (*luMatrix)[r1] = (*luMatrix)[r2];
    (*luMatrix)[r2] = *tmpM;
    tmp = (*prmVector)[r1];                // save the permutation info
    (*prmVector)[r1] = (*prmVector)[r2];
    (*prmVector)[r2] = tmp;
    permutationParity = -permutationParity;

    delete tmpM;
}

void matrix::initialiseLUMatrices(void) {
    luMatrix = new vector<vector<T> >;
    luMatrix->resize(rowNo);
	for(unsigned int r=0; r<=rowNo-1; ++r)
		(*luMatrix)[r].resize(colNo);
    prmVector = new vector<unsigned int>;
    prmVector->resize(rowNo);
    sclVector = new vector<T>;
    sclVector->resize(rowNo);
}

void matrix::deleteLUMatrices(void) {
	delete luMatrix;
    delete prmVector;
	delete sclVector;
	luMatrix = NULL;
	prmVector = NULL;
	sclVector = NULL;
	permutationParity = 1;
}

void matrix::copyLUMatrices(const matrix & m) {
    unsigned int rowSize = colNo * sizeof(T);

    if (m.luMatrix) {
        luMatrix = new vector<vector<T> >;
        prmVector = new vector<unsigned int>;
        sclVector = new vector<T>;
        *luMatrix = *m.luMatrix;
		*prmVector = *m.prmVector;
		*sclVector = *m.sclVector;
        permutationParity = m.permutationParity;
    } else {
 	    luMatrix = NULL;
        prmVector = NULL;
        sclVector = NULL;
        permutationParity = 1;
    }
}


// ***************************************************************
// ******************* NON-MODYFYING OPERATORS *******************
// **********(not member functions of the matrix class)***********
// ***************************************************************

// unary minus operator
matrix operator- (const matrix & m) {
    return m*(-1);
}

// binary addition operator
matrix operator+ (const matrix & m1, const matrix & m2) {
    matrix tmp = m1;
    tmp += m2;
    return tmp;
}

// scalar binary multiplication operator (matrix * scalar)
matrix operator* (const matrix & m, const T & v) {
    matrix tmp = m;
    tmp *= v;
    return tmp;
}

// scalar binary multiplication operator (scalar * matrix)
matrix operator* (const T & v, const matrix & m) {
    return m*v;
}

// matrix binary multiplication operator
matrix operator* (const matrix & m1, const matrix & m2) {
    matrix tmp = m1;
    tmp *= m2;
    return tmp;
}

// output stream function
ostream & operator << (ostream & ostr, const matrix & m)
{
    for(unsigned int r=0; r<=m.getRowNo()-1; ++r) {
        for(unsigned int c=0; c<=m.getColNo()-1; ++c) {
            ostr << m(r,c) << '\t';
        }
        ostr << endl;
    }
    return ostr;
}

} // namespace std
