#ifndef MATRIX_H
#define MATRIX_H

#include <iostream>
#include <vector>
#include <cmath>

namespace std {

typedef float T;

class matrix {
public:
    // constructors
    matrix(unsigned int rowNo, unsigned int colNo = 1, T v = 0);
    matrix(const matrix& m);   // copy constructor

    // destructor
    ~matrix ();

    // few useful functions
    matrix &transpose();
    matrix &randomize(unsigned int limit = 10);
    unsigned int getRowNo() const { return rowNo; }
    unsigned int getColNo() const { return colNo; }

    // overloaded self-modifying operators
    matrix& operator= (const matrix& m);
    matrix& operator+= (const matrix& m);
    matrix& operator*= (const matrix& m);
    matrix& operator*= (const T& v);

    // overloaded comparison operators
    bool operator== (const matrix & m) const {return (*vals == *m.vals);}
    bool operator!= (const matrix & m) const {return !operator==(m);}

    // overloaded subscript operators (for modification & access)
    T& operator()(unsigned int r, unsigned int c = 0);
    T  operator()(unsigned int r, unsigned int c = 0) const;

    // access to LU-related data
    unsigned int getLUPermutationVector(unsigned int r);
    T getLUScalingVector(unsigned int r);
    matrix getLUPermutationVector(void);
    matrix getLUScalingVector(void);
    matrix getLMatrix(void);
    matrix getUMatrix(void);

    // LS solving
    matrix solveLS(const matrix& b);
    
    // the determinant
    T det(void);
    
private:
    unsigned int rowNo, colNo;
    vector<vector<T> > * vals, *luMatrix;   
    vector<T> * sclVector;
    vector<unsigned int> * prmVector;
    short int permutationParity;

    void error(const char* pMsg) const;
    
    void luDecompose(void);
    void luRescale(void);
    void luSwapRows(unsigned int r1, unsigned int r2);
    void initialiseLUMatrices(void);
    void deleteLUMatrices(void);
    void copyLUMatrices(const matrix & m);
};

// ******************* NON-MODYFYING OPERATORS *******************

matrix operator- (const matrix & m);
matrix operator+ (const matrix & m1, const matrix & m2);
matrix operator* (const matrix & m, const T & v);
matrix operator* (const T & v, const matrix & m);
matrix operator* (const matrix & m1, const matrix & m2);
ostream & operator << (ostream & ostr, const matrix & m);

} // namespace std

#endif // MATRIX_H
